;
; free-dictionary.el
;
; Emacs lisp routine to look up a word from the the 'Free Dictionary'
; (http://thefreedictionary.com)
;
;
; This project was sponsored by the
; Open Source Financial Developers Association (OSFDA)
;
;      http://osfda.org
;
; A download _might_ be available from http://osfda.org/downloads
;
;
; LICENSE: GPL2
;
; This work is free software; you can redistribute it and/or modify it
; under the terms of the GNU General Public License as published
; by the Free Software Foundation; either version 2 of the License, or
; any later version. This work is distributed in the hope that it will be
; useful, but without any warranty; without even the implied warranty
; of merchantability or fitness for a particular purpose.
;

(setq warned-changed)

(defun backward-line(&optional count)
       (if (not count)
           (setq count 1)
       )
       (forward-line (- 0 count))
) ; backward-line(&optional count)

(defun map-from-html-entities() (interactive)
       (let ((entities '(("&amp;" "&")
                         ("&nbsp;" " ")
                         ("&lt;" "<")
                         ("&gt;" ">")
                         ("&quot;" "\"")
                         ("&apos;" "'")
                         ("&copy;" "(c)")
                         ("&reg;" "(r)")
                         ("&ndash;" "-")
                         ("&mdash;" "--")
                         ("&ldquo;" "\"")
                         ("&rdquo;" "\"")
                         ("&lsquo;" "'")
                         ("&rsquo;" "'")
                        )
             )
            )
            (dolist (entity entities)
                    (goto-char (point-min))
                    (replace-string (nth 0 entity) (nth 1 entity))
            )
       ) ; let

       (goto-char (point-min))
       (replace-regexp "&#x[a-f0-9]+;" "")

       (goto-char (point-min))
       (replace-regexp "&#[0-9]+;" "")

) ; map-from-html-entities()

(defun remove-script-blocks()
       (goto-char (point-min))
       (let ((NO-BOUNDS nil)
             (NO-FAIL t)
            )
            (while (search-forward-regexp "<script[[:space:]>]"
                                          NO-BOUNDS
                                          NO-FAIL
                   ) ; more reliable than a multiline regexp...
                   (replace-match "")
                   (set-mark (point))
                   (search-forward "</script>")
                   (kill-region (mark) (point))
                   (deactivate-mark)
            ) ; while...
       ) ; let...
       (goto-char (point-min))
) ; remove-script-blocks()

(defun remove-html-tags()
; Note that this is NOT a formal HTML parser!
; There are many cases of markup syntax that this approach could fail with.
; But given that _to date_ the markup generated by The Free Dictionary site
; has not been (mischeviously?) crafted that way, this suffices.
; In deference to The Free Dictionary making their service available,
; I have purposely left all their copy in place
; (including any advertising one-liners...)
       (goto-char (point-min))
       (replace-regexp "<.*?>" " ") ; tags...

       (goto-char (point-min))
) ; remove-html-tags()

(defun map-breaks-to-newlines()
       (goto-char (point-min))
       (replace-string "<br>" "\n\n\n")

       (goto-char (point-min))
       (replace-string "<br/>" "\n\n\n")

       (goto-char (point-min))
       (replace-string "<br />" "\n\n\n")

       (goto-char (point-min))
) ; map-breaks-to-newlines()

(defun map-glyphs()
 (let ((glyphs
        '(("&#x2192;" "=--&gt;") ; =-->

          ;; the two leading spaces here before the "!" are to override
          ;; later interpretation of "!" as trailing punctuation
          ;; (see fixup-spacing...)
          ("<span class=Ant>" "  != ")

          ("<div class=Rel>" "&lt;=&gt;") ; <=>
         )
       )
      )
      (dolist (glyph glyphs)
              (goto-char (point-min))
              (replace-string (nth 0 glyph) (nth 1 glyph))
      )
 )
) ; map-glyphs()

(defun strip-html() (interactive)
       (replace-string "<h2>" "\n\n\n")
       (map-breaks-to-newlines)
       (replace-string "</option>" "; ")
       (map-glyphs)
       (remove-html-tags)
       (map-from-html-entities)
) ; strip-html()

(defun strip-carriage-returns()
       (goto-char (point-min))
       (replace-string "\r" "") ; appear as ^M's
) ;  strip-carriage-returns()

(defun remove-unicode()
       ;; possible improvement: replace this with fold-unicode,
       ;; which converts accented characters to extended ascii ones...
       (goto-char (point-min))
       (replace-regexp "[^[:ascii:]]+" "")
) ; remove-unicode()

(defun delete-multiple-blank-lines() (interactive) ; (to a single one...)
       (delete-trailing-whitespace)
       (goto-char (point-min))
       (while (re-search-forward "\n\n\n" nil t)
              (backward-line)
              (delete-blank-lines)
       )
) ; delete-multiple-blank-lines()

(defun delete-leading-spaces() (interactive)
       (goto-char (point-min))
       (replace-regexp "^[ \t]+" "")
) ; delete-multiple-spaces()

(defun delete-multiple-spaces() (interactive)
       (goto-char (point-min))
       (replace-regexp "[ \t]+" " ")
) ; delete-multiple-spaces()

(defun fixup-spacing() (interactive)
       (delete-leading-spaces)
       (delete-multiple-spaces)
       (let ((leading-punctuation "([{'")
             (trailing-punctuation ",.;)]}?!'")
            )
            (dotimes (offset (length trailing-punctuation))
                     (goto-char (point-min))
                     (let ((trailer
                           (substring trailing-punctuation
                                      offset (+ offset 1)))
                          )
                          (replace-string (concat " " trailer) trailer)
                     )
            )
            (dotimes (offset (length leading-punctuation))
                     (goto-char (point-min))
                     (let ((leader
                           (substring leading-punctuation
                                      offset (+ offset 1)))
                          )
                          (replace-string (concat leader " ") leader)
                     )
            )
       ) ; let...
       (goto-char (point-min))
) ; fixup-spacing()

(defun adjust-layout()
; returns #layouts that could not be found (so #changes...)
 (let ((layouts
        '(("\nTranslations\n" "\n\nTranslations:\n")
          ("You can also log in with\n\nFacebook\nTwitter\nGoogle+\nYahoo\n"
           "You can also log in with: Facebook Twitter Google+ Yahoo\n"
          )
          ("Share:\n\nFacebook\nTwitter\nLinkedIn\nMail\n"
           "Share: Facebook Twitter LinkedIn Mail\n"
          )
          ("Follow:\n\nFacebook\nTwitter\nGoogle+\nRss\nMail\n"
           "Follow: Facebook Twitter Google+ Rss Mail\n"
          )
          ("Link to this page:\n\nFacebook\nTwitter\n"
           "Link to this page: Facebook Twitter\n"
          )
          ("\nDisclaimer\n"
           "\nDisclaimer:\n"
          )
         )  
       )
       (changes 0)
      )
      (dolist (layout layouts)
              (goto-char (point-min))
              (replace-string (nth 0 layout) (nth 1 layout))
              (if (not (> (point) (point-min)))
                  (setq changes (+ changes 1))
              )
      )
      (goto-char (point-min))
      changes
 ) ; let ((layouts...
) ; adjust-layout()

(defun free-dictionary-fixup-markup()
       (strip-carriage-returns)
       (remove-unicode)
       (remove-script-blocks)

       ;; first h2 is where the result is; remove all the stuff before it...
       (set-mark (point-min))
       (let* ((NO-BOUNDS nil)
              (NO-FAIL t)
              (SUGGESTION-PROMPT "Did you mean:")
              (found (search-forward "<h2>" NO-BOUNDS NO-FAIL))
              (suggestions)
              (OTHER-SUGGESTION "Can't find what you are looking for?")
              (not-changed)
             )
             (when (not found)
                   (goto-char (point-min))
                   (setq suggestions
                         (search-forward SUGGESTION-PROMPT
                                         NO-BOUNDS
                                         NO-FAIL
                         )
                   )
                  (if suggestions
                      (backward-char (length SUGGESTION-PROMPT))
                  )
             ) ; when (not found)
             (if (or found suggestions)
                 (progn ; safer than a multiline regexp...
                  (kill-region (mark) (point))
                  (deactivate-mark)

                  (strip-html)
;;                  (fill-region (point-min) (point-max))
                  (fixup-spacing)
                  (delete-multiple-blank-lines)

                  (goto-char (point-min))
                  (replace-string "\n?" ":")

                  ;; have each numbered section begin on a new line...
                  (goto-char (point-min))
                  (replace-regexp "\\([0-9]+\\)\\." "\n\\1.")

                  ;; have each lettered section begin on a new line...
                  (goto-char (point-min))
                  (let ((case-fold-search-setting case-fold-search))
                       (setq case-fold-search nil)
                       (replace-regexp "\\([^a-z]\\)\\([a-m]\\)\\."
                                       "\\1\n\\2."
                       )
                       (setq case-fold-search case-fold-search-setting)
                  )
                  ;; have each lettered section begin indented...
                  (goto-char (point-min))
                  ;; tried a tab character for indentation,
                  ;; but it was too much;
                  ;; for now, we'll just use three blanks...
                  (replace-regexp "^\\([a-m]\\)\\." "   \\1.")

                  (setq not-changed
                        (= (adjust-layout)
                           0
                        )
                  )

                  (if suggestions
                      (progn
                       (replace-string (concat OTHER-SUGGESTION
                                               " "
                                       )
                                       (concat OTHER-SUGGESTION
                                               "\n"
                                       )
                       )
                       (goto-char (point-min))
                       (forward-line)
                       (forward-line)
                      ) ; if suggestions
                      (forward-line) ; else (not suggestions)
                  ) ; if suggestions
                  (visual-line-mode)
                  (if (or not-changed
                          warned-changed ; they already have been warned...
                      )
                      (message "") ; conceal editing status messages...
                      (progn ; else
                       (message "warning: The Free Dictionary's format has changed; this macro needs updating...")
                       (setq warned-changed t)
                      ) ; progn
                  ) ; if...
                 ) ; (or (search-forward "<h2>"...
                 (progn ; else (not (or found suggestions))
                  (delete-window)
                  (message "word not found...")
                 )
             ) ; if (or found suggestions)
        ) ; let...
) ; free-dictionary-fixup-markup()

(defun free-dictionary(&optional word) (interactive)
"
Looks up a word in the 'Free Dictionary' (http://thefreedictionary.com);
this requires an Internet connection.

A word can be passed as an argument; if omitted, the word will be retrieved
from the current word that the cursor is on. If the cursor is not on a word,
the end-user will be prompted for one.

For now, conversion of the web-based markup into a text format
is done on a 'best effort' basis; fortunately, the looked-up info
should all be there.

(Unicode character support, such as support for accented characters,
 has not yet been implemented; Unicode characters are removed for now...)
"
 ;;
 ;; ISSUES:
 ;;
 ;; 1) after using the standard function url-retrieve,
 ;;    screen updating can sometimes end up out of sync (!)
 ;;    To try and remedy this, a redraw-display is presently done
 ;;    at the end; but unfortunately, it seems that sometimes
 ;;    you can still get out of sync at some later point...
 ;;
 ;; 2) Unicode characters are not yet implemented; mapping these to support
 ;;    accented characters, etc. would be possible, and would make for a
 ;;    nice minor improvement...
 ;;

 (let* ((FREE-DICTIONARY-HOST-NAME
         "thefreedictionary.com"
        )
        (FREE-DICTIONARY-URI
         (concat "http://www."
                 FREE-DICTIONARY-HOST-NAME
         )
        )
        (CONNECTION-ERROR-MESSAGE
         (concat "error connecting to "
                 FREE-DICTIONARY-HOST-NAME
                 "..."
         )
        )
        (WORD-STRICT t)
        (REALLY-WORD t)
        (PROMPT "word: ")
       )
       (defconst FREE-DICTIONARY-BUFFER-NAME-PATTERN
                 (concat "^ \\*http .*"
                         FREE-DICTIONARY-HOST-NAME
                         ":"
                 )
       ) ; this needs to be referenced by an asynchronous event macro...
       (when (not word) ; wasn't passed as an argument...
             ; try and get it from the cursor point...
             (setq word (current-word WORD-STRICT REALLY-WORD))
             (when (not word) ; still don't have it? prompt for it...
                   (setq word (read-from-minibuffer PROMPT))
             ) ; when (not word)
       ) ; when (not word)
       (when (> (length word) 0)
             (url-retrieve (concat FREE-DICTIONARY-URI "/" word)
                           (lambda (status)
                                   (if status
                                       (message CONNECTION-ERROR-MESSAGE)
                                       (progn ; else status===nil
                                        (let ((free-dictionary-buffer
                                               (current-buffer)
                                              )
                                              (window
                                               (catch 'dolist-window
                                                      (dolist (window (window-list))
                                                              (if (string-match FREE-DICTIONARY-BUFFER-NAME-PATTERN
                                                                                (buffer-name (window-buffer window))
                                                                  )
                                                                  (throw 'dolist-window
                                                                         window
                                                                  )
                                                              )
                                                      )
                                                      nil
                                               )
                                              )
                                             )

                                             (if window
                                                 (select-window window)
                                                 (when (not (string-match FREE-DICTIONARY-BUFFER-NAME-PATTERN
                                                                          (buffer-name (window-buffer))
                                                            )
                                                       )
                                                       (split-window)
                                                       (select-window (next-window))
                                                 )
                                             ) ; if window
                                             (switch-to-buffer free-dictionary-buffer)
                                        ) ; let ((free-dictionary-buffer...
                                        (free-dictionary-fixup-markup)

                                        ;; url-retrieve (which uses curl?)
                                        ;; appears to have a side effect
                                        ;; on the display; if the following is
                                        ;; not shown over time to be a
                                        ;; stable fix, then we need some sort
                                        ;; of setterm or other little-known
                                        ;; emacs reset command; people familiar
                                        ;; with the internals of emacs curses
                                        ;; addressing will need to advise...
                                        (redraw-display)
                                       ) ; else status===nil
                                   ) ; if status...
                           ) ; lambda (status)
             ) ; url-retrieve...
       ) ; when (> (length word) 0)
 ) ; let*
) ; free-dictionary(&optional word)
